﻿// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Security.Claims;
using Microsoft.Identity.Client.Utils;
using Microsoft.Identity.Json;
using Microsoft.Identity.Json.Linq;

namespace Microsoft.Identity.Client.Internal
{
    internal static class IdTokenClaim
    {
        public const string Issuer = "iss";
        public const string ObjectId = "oid";
        public const string Subject = "sub";
        public const string TenantId = "tid";
        public const string Version = "ver";
        public const string PreferredUsername = "preferred_username";
        public const string Name = "name";
        public const string HomeObjectId = "home_oid";
        public const string GivenName = "given_name";
        public const string FamilyName = "family_name";
        public const string Upn = "upn";
    }

    internal static class JsonClaimValueTypes
    {
        public const string Json = "JSON";
        public const string JsonArray = "JSON_ARRAY";
        public const string JsonNull = "JSON_NULL";
    }

    internal class IdToken
    {
        private const string DefaultIssuser = "LOCAL AUTHORITY";

        public string ObjectId { get; private set; }

        public string Subject { get; private set; }

        public string TenantId { get; private set; }

        public string PreferredUsername { get; private set; }

        public string Name { get; private set; }

        public string Upn { get; private set; }

        public string GivenName { get; private set; }

        public string FamilyName { get; private set; }

        public string GetUniqueId()
        {
            return ObjectId ?? Subject;
        }

        public ClaimsPrincipal ClaimsPrincipal { get; private set; }

        public static IdToken Parse(string idToken)
        {
            if (string.IsNullOrEmpty(idToken))
            {
                return null;
            }

            string[] idTokenSegments = idToken.Split(new[] { '.' });

            if (idTokenSegments.Length < 2)
            {
                throw new MsalClientException(
                    MsalError.InvalidJwtError,
                    MsalErrorMessage.IDTokenMustHaveTwoParts);
            }

            try
            {
                string payload = Base64UrlHelpers.Decode(idTokenSegments[1]);
                var idTokenClaims = JsonConvert.DeserializeObject<Dictionary<string, object>>(payload);                

                IdToken parsedIdToken = new IdToken();

                List<Claim> claims = GetClaimsFromRawToken(idTokenClaims);
                parsedIdToken.ClaimsPrincipal = new ClaimsPrincipal(new ClaimsIdentity(claims));

                parsedIdToken.ObjectId = parsedIdToken.ClaimsPrincipal.FindFirst(IdTokenClaim.ObjectId)?.Value;
                parsedIdToken.Subject = parsedIdToken.ClaimsPrincipal.FindFirst(IdTokenClaim.Subject)?.Value;
                parsedIdToken.TenantId = parsedIdToken.ClaimsPrincipal.FindFirst(IdTokenClaim.TenantId)?.Value;
                parsedIdToken.PreferredUsername = parsedIdToken.ClaimsPrincipal.FindFirst(IdTokenClaim.PreferredUsername)?.Value;
                parsedIdToken.Name = parsedIdToken.ClaimsPrincipal.FindFirst(IdTokenClaim.Name)?.Value;
                parsedIdToken.Upn = parsedIdToken.ClaimsPrincipal.FindFirst(IdTokenClaim.Upn)?.Value;
                parsedIdToken.GivenName = parsedIdToken.ClaimsPrincipal.FindFirst(IdTokenClaim.GivenName)?.Value;
                parsedIdToken.FamilyName = parsedIdToken.ClaimsPrincipal.FindFirst(IdTokenClaim.FamilyName)?.Value;

                return parsedIdToken;

            }
            catch (JsonException exc)
            {
                throw new MsalClientException(
                    MsalError.JsonParseError,
                    MsalErrorMessage.FailedToParseIDToken,
                    exc);
            }
        }

        #region IdToken to Claims parsing - logic copied from Wilson!
        private static List<Claim> GetClaimsFromRawToken(Dictionary<string, object> idTokenClaims)
        {
            List<Claim> claims = new List<Claim>();

            string issuer = null;
            if (idTokenClaims.TryGetValue(IdTokenClaim.Issuer, out object issuerObj))
            {
                issuer = issuerObj as string;                
            }
            issuer = issuer ?? DefaultIssuser;

            foreach (KeyValuePair<string, object> keyValuePair in idTokenClaims)
            {
                if (keyValuePair.Value == null)
                {
                    claims.Add(new Claim(keyValuePair.Key, string.Empty, JsonClaimValueTypes.JsonNull, issuer, issuer));
                    continue;
                }

                var claimValue = keyValuePair.Value as string;
                if (claimValue != null)
                {
                    claims.Add(new Claim(keyValuePair.Key, claimValue, ClaimValueTypes.String, issuer, issuer));
                    continue;
                }

                var jtoken = keyValuePair.Value as JToken;
                if (jtoken != null)
                {
                    AddClaimsFromJToken(claims, keyValuePair.Key, jtoken, issuer);
                    continue;
                }

                var objects = keyValuePair.Value as IEnumerable<object>;
                if (objects != null)
                {
                    foreach (var obj in objects)
                    {
                        claimValue = obj as string;
                        if (claimValue != null)
                        {
                            claims.Add(new Claim(keyValuePair.Key, claimValue, ClaimValueTypes.String, issuer, issuer));
                            continue;
                        }

                        jtoken = obj as JToken;
                        if (jtoken != null)
                        {
                            AddDefaultClaimFromJToken(claims, keyValuePair.Key, jtoken, issuer);
                            continue;
                        }

                        // DateTime claims require special processing. JsonConvert.SerializeObject(obj) will result in "\"dateTimeValue\"". The quotes will be added.
                        if (obj is DateTime dateTimeValue)
                            claims.Add(
                                new Claim(
                                    keyValuePair.Key, 
                                    dateTimeValue.ToUniversalTime().ToString("o", CultureInfo.InvariantCulture), 
                                    ClaimValueTypes.DateTime, 
                                    issuer, 
                                    issuer));
                        else
                            claims.Add(
                                new Claim(
                                    keyValuePair.Key, 
                                    JsonConvert.SerializeObject(obj),
                                    GetClaimValueType(obj), 
                                    issuer, 
                                    issuer));
                    }

                    continue;
                }

                IDictionary<string, object> dictionary = keyValuePair.Value as IDictionary<string, object>;
                if (dictionary != null)
                {
                    foreach (var item in dictionary)
                        claims.Add(new Claim(keyValuePair.Key, "{" + item.Key + ":" + JsonConvert.SerializeObject(item.Value) + "}", GetClaimValueType(item.Value), issuer, issuer));

                    continue;
                }

                // DateTime claims require special processing. JsonConvert.SerializeObject(keyValuePair.Value) will result in "\"dateTimeValue\"". The quotes will be added.
                if (keyValuePair.Value is DateTime dateTime)
                    claims.Add(new Claim(keyValuePair.Key, dateTime.ToUniversalTime().ToString("o", CultureInfo.InvariantCulture), ClaimValueTypes.DateTime, issuer, issuer));
                else
                    claims.Add(new Claim(keyValuePair.Key, JsonConvert.SerializeObject(keyValuePair.Value), GetClaimValueType(keyValuePair.Value), issuer, issuer));
            }

            return claims;
        }
    
        private static void AddClaimsFromJToken(List<Claim> claims, string claimType, JToken jtoken, string issuer)
        {
            if (jtoken.Type == JTokenType.Object)
            {
                claims.Add(new Claim(claimType, jtoken.ToString(Formatting.None), JsonClaimValueTypes.Json));
            }
            else if (jtoken.Type == JTokenType.Array)
            {
                var jarray = jtoken as JArray;
                foreach (var item in jarray)
                {
                    switch (item.Type)
                    {
                        case JTokenType.Object:
                            claims.Add(new Claim(claimType, item.ToString(Formatting.None), JsonClaimValueTypes.Json, issuer, issuer));
                            break;

                        // only go one level deep on arrays.
                        case JTokenType.Array:
                            claims.Add(new Claim(claimType, item.ToString(Formatting.None), JsonClaimValueTypes.JsonArray, issuer, issuer));
                            break;

                        default:
                            AddDefaultClaimFromJToken(claims, claimType, item, issuer);
                            break;
                    }
                }
            }
            else
            {
                AddDefaultClaimFromJToken(claims, claimType, jtoken, issuer);
            }
        }

        private static void AddDefaultClaimFromJToken(List<Claim> claims, string claimType, JToken jtoken, string issuer)
        {
            JValue jvalue = jtoken as JValue;
            if (jvalue != null)
            {
                // String is special because item.ToString(Formatting.None) will result in "/"string/"". The quotes will be added.
                // Boolean needs item.ToString otherwise 'true' => 'True'
                if (jvalue.Type == JTokenType.String)
                    claims.Add(new Claim(claimType, jvalue.Value.ToString(), ClaimValueTypes.String, issuer, issuer));
                // DateTime claims require special processing. jtoken.ToString(Formatting.None) will result in "\"dateTimeValue\"". The quotes will be added.
                else if (jvalue.Value is DateTime dateTimeValue)
                    claims.Add(new Claim(claimType, dateTimeValue.ToUniversalTime().ToString("o", CultureInfo.InvariantCulture), ClaimValueTypes.DateTime, issuer, issuer));
                else
                    claims.Add(new Claim(claimType, jtoken.ToString(Formatting.None), GetClaimValueType(jvalue.Value), issuer, issuer));
            }
            else
                claims.Add(new Claim(claimType, jtoken.ToString(Formatting.None), GetClaimValueType(jtoken), issuer, issuer));
        }

        private static string GetClaimValueType(object obj)
        {
            if (obj == null)
                return JsonClaimValueTypes.JsonNull;

            var objType = obj.GetType();

            if (objType == typeof(string))
                return ClaimValueTypes.String;

            if (objType == typeof(int))
                return ClaimValueTypes.Integer;

            if (objType == typeof(bool))
                return ClaimValueTypes.Boolean;

            if (objType == typeof(double))
                return ClaimValueTypes.Double;

            if (objType == typeof(long))
            {
                long l = (long)obj;
                if (l >= int.MinValue && l <= int.MaxValue)
                    return ClaimValueTypes.Integer;

                return ClaimValueTypes.Integer64;
            }

            if (objType == typeof(DateTime))
                return ClaimValueTypes.DateTime;

            if (objType == typeof(JObject))
                return JsonClaimValueTypes.Json;

            if (objType == typeof(JArray))
                return JsonClaimValueTypes.JsonArray;

            return objType.ToString();
        }
        
        #endregion
    }
}
