﻿// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

using System;
using System.Drawing.Text;
using System.IO;
using System.Security.Cryptography.X509Certificates;
using System.Security.Permissions;
using Microsoft.Identity.Client;
using Microsoft.Identity.Client.Internal;
using Microsoft.Identity.Test.Common;
using Microsoft.Identity.Test.Common.Core.Helpers;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace Microsoft.Identity.Test.Unit.AppConfigTests
{
    [TestClass]
    [TestCategory("BuilderTests")]
    public class ConfidentialClientApplicationBuilderTests
    {
        [TestInitialize]
        public void TestInitialize()
        {
            TestCommon.ResetInternalStaticCaches();
        }

        [TestMethod]
        public void TestConstructor()
        {
            var cca = ConfidentialClientApplicationBuilder.Create(TestConstants.ClientId)
                                                          .WithClientSecret("cats")
                                                          .Build();

            Assert.AreEqual(TestConstants.ClientId, cca.AppConfig.ClientId);
            Assert.IsNotNull(cca.UserTokenCache);

            // Validate Defaults
            Assert.AreEqual(LogLevel.Info, cca.AppConfig.LogLevel);
            Assert.AreEqual(TestConstants.ClientId, cca.AppConfig.ClientId);
            Assert.IsNotNull(cca.AppConfig.ClientName);
            Assert.IsNotNull(cca.AppConfig.ClientVersion);
            Assert.AreEqual(false, cca.AppConfig.EnablePiiLogging);
            Assert.IsNull(cca.AppConfig.HttpClientFactory);
            Assert.AreEqual(false, cca.AppConfig.IsDefaultPlatformLoggingEnabled);
            Assert.IsNull(cca.AppConfig.LoggingCallback);
            Assert.AreEqual(Constants.DefaultConfidentialClientRedirectUri, cca.AppConfig.RedirectUri);
            Assert.AreEqual(null, cca.AppConfig.TenantId);
        }

        private ConfidentialClientApplicationOptions CreateConfidentialClientApplicationOptions()
        {
            return new ConfidentialClientApplicationOptions
            {
                ClientId = TestConstants.ClientId,
                ClientSecret = "the_client_secret",
                TenantId = "the_tenant_id",
            };
        }

        private void TestBuildConfidentialClientFromOptions(ConfidentialClientApplicationOptions options)
        {
            options.ClientSecret = "cats";
            var app = ConfidentialClientApplicationBuilder.CreateWithApplicationOptions(options).Build();
            var authorityInfo = ((ConfidentialClientApplication)app).ServiceBundle.Config.Authority.AuthorityInfo;
            Assert.AreEqual("https://login.microsoftonline.com/the_tenant_id/", authorityInfo.CanonicalAuthority);
        }

        [TestMethod]
        [DeploymentItem(@"Resources\testCert.crtfile")]
        public void TestBuildWithNoClientSecretButUsingCert()
        {
            var options = new ConfidentialClientApplicationOptions()
            {
                ClientId = TestConstants.ClientId,
                TenantId = "the_tenant_id",
                Instance = "https://login.microsoftonline.com",
            };

            var cert = new X509Certificate2(
               ResourceHelper.GetTestResourceRelativePath("testCert.crtfile"), "passw0rd!");

            var app = ConfidentialClientApplicationBuilder.CreateWithApplicationOptions(options)
                                                          .WithCertificate(cert)
                                                          .Build();
            var authority = ((ConfidentialClientApplication)app).ServiceBundle.Config.Authority;
            Assert.AreEqual("https://login.microsoftonline.com/the_tenant_id/", authority.AuthorityInfo.CanonicalAuthority);
        }

        [TestMethod]
        public void TestBuildWithInstanceWithTrailingSlash()
        {
            var options = CreateConfidentialClientApplicationOptions();
            options.Instance = "https://login.microsoftonline.com/";
            TestBuildConfidentialClientFromOptions(options);
        }

        [TestMethod]
        public void CacheSynchronizationWithDefaultCCA()
        {
            //Validate CacheSynchronizationEnabled when app is created from ApplicaitonOptions for CCA
            var options = new ConfidentialClientApplicationOptions()
            {
                ClientSecret = "secret",
                ClientId = TestConstants.ClientId,
            };
            var app = ConfidentialClientApplicationBuilder.CreateWithApplicationOptions(options).Build();
            Assert.IsFalse((app.AppConfig as ApplicationConfiguration).CacheSynchronizationEnabled);

            options = new ConfidentialClientApplicationOptions
            {
                ClientId = TestConstants.ClientId,
                ClientSecret = "secret",
                EnableCacheSynchronization = false
            };
            app = ConfidentialClientApplicationBuilder.CreateWithApplicationOptions(options).Build();
            Assert.AreEqual(false, (app.AppConfig as ApplicationConfiguration).CacheSynchronizationEnabled);

            options = new ConfidentialClientApplicationOptions
            {
                ClientId = TestConstants.ClientId,
                ClientSecret = "secret",
                EnableCacheSynchronization = true
            };
            app = ConfidentialClientApplicationBuilder.CreateWithApplicationOptions(options).Build();
            Assert.AreEqual(true, (app.AppConfig as ApplicationConfiguration).CacheSynchronizationEnabled);

            //Validate CacheSynchronizationEnabled is false by default when app is created from ConfidentialClientApplicationBuilder
            app = ConfidentialClientApplicationBuilder.Create(Guid.NewGuid().ToString()).WithClientSecret(TestConstants.ClientSecret).BuildConcrete();
            Assert.IsFalse((app.AppConfig as ApplicationConfiguration).CacheSynchronizationEnabled);

            //Validate CacheSynchronizationEnabled when app is created from ApplicaitonOptions for PCA
            var options2 = new PublicClientApplicationOptions()
            {
                ClientId = TestConstants.ClientId
            };
            var app2 = PublicClientApplicationBuilder.CreateWithApplicationOptions(options2).Build();
            Assert.IsTrue((app2.AppConfig as ApplicationConfiguration).CacheSynchronizationEnabled);

            //Validate CacheSynchronizationEnabled is true by default when app is created from PublicClientApplicationBuilder
            app2 = PublicClientApplicationBuilder.Create(Guid.NewGuid().ToString()).BuildConcrete();
            Assert.IsTrue((app2.AppConfig as ApplicationConfiguration).CacheSynchronizationEnabled);
        }

        [DataTestMethod]
        [DataRow(false, false, false)]
        [DataRow(true, true, true)]
        [DataRow(true, false, false)]
        [DataRow(false, true, true)]        
        public void CacheSynchronizationNoDefault(bool optionFlag, bool builderFlag, bool result)
        {
            var options = new ConfidentialClientApplicationOptions
            {
                ClientSecret = "secret",
                ClientId = TestConstants.ClientId,
                EnableCacheSynchronization = optionFlag
            };
            var app = ConfidentialClientApplicationBuilder.CreateWithApplicationOptions(options).WithCacheSynchronization(builderFlag).Build();
            Assert.AreEqual(result, (app.AppConfig as ApplicationConfiguration).CacheSynchronizationEnabled);
        }

        [TestMethod]
        public void TestBuildWithInstanceWithoutTrailingSlash()
        {
            var options = CreateConfidentialClientApplicationOptions();
            options.Instance = "https://login.microsoftonline.com";
            TestBuildConfidentialClientFromOptions(options);
        }

        [TestMethod]
        public void TestBuildWithNullInstance()
        {
            var options = CreateConfidentialClientApplicationOptions();
            options.Instance = null;
            TestBuildConfidentialClientFromOptions(options);
        }

        [TestMethod]
        public void TestBuildWithEmptyInstance()
        {
            var options = CreateConfidentialClientApplicationOptions();
            options.Instance = string.Empty;
            TestBuildConfidentialClientFromOptions(options);
        }

        [TestMethod]
        public void TestWithDifferentClientId()
        {
            const string ClientId = "9340c42a-f5de-4a80-aea0-874adc2ca325";
            const string AppIdUri = "https://microsoft.onmicrosoft.com/aa3e634f-58b3-4eb7-b4ed-244c44c29c47";
            var cca = ConfidentialClientApplicationBuilder.Create(ClientId).WithClientSecret("cats").Build();
            Assert.AreEqual(ClientId, cca.AppConfig.ClientId);

            cca = ConfidentialClientApplicationBuilder.Create(AppIdUri).WithClientSecret("cats").Build();
            Assert.AreEqual(AppIdUri, cca.AppConfig.ClientId);
        }

        [TestMethod]
        public void TestConstructor_ClientIdOverride()
        {
            const string ClientId = "73cc145e-798f-430c-8d6d-618f1a5802e9";
            var cca = ConfidentialClientApplicationBuilder.Create(TestConstants.ClientId)
                                                                .WithClientId(ClientId)
                                                                .WithClientSecret("cats")
                                                                .Build();
            Assert.AreEqual(ClientId, cca.AppConfig.ClientId);
        }

        [TestMethod]
        public void TestConstructor_WithClientNameAndVersion()
        {
            const string ClientName = "my client name";
            const string ClientVersion = "1.2.3.4-prerelease";
            var cca =
                ConfidentialClientApplicationBuilder.Create(TestConstants.ClientId)
                                                            .WithClientName(ClientName)
                                                            .WithClientVersion(ClientVersion)
                                                            .WithClientSecret("cats")
                                                            .Build();
            Assert.AreEqual(ClientName, cca.AppConfig.ClientName);
            Assert.AreEqual(ClientVersion, cca.AppConfig.ClientVersion);
        }

        [TestMethod]
        public void TestConstructor_WithDebugLoggingCallback()
        {
            var cca = ConfidentialClientApplicationBuilder.Create(TestConstants.ClientId)
                                                                .WithClientSecret("cats")
                                                                .WithDebugLoggingCallback()
                                                                .Build();
            Assert.IsNotNull(cca.AppConfig.LoggingCallback);
        }

        [TestMethod]
        public void TestConstructor_WithHttpClientFactory()
        {
            var httpClientFactory = NSubstitute.Substitute.For<IMsalHttpClientFactory>();
            var cca = ConfidentialClientApplicationBuilder.Create(TestConstants.ClientId)
                                                                .WithHttpClientFactory(httpClientFactory)
                                                                .WithClientSecret("cats")
                                                                .Build();
            Assert.AreEqual(httpClientFactory, cca.AppConfig.HttpClientFactory);
        }

        [TestMethod]
        public void TestConstructor_WithLogging()
        {
            var cca = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                            .WithClientSecret("cats")
                            .WithLogging((level, message, pii) => { }).Build();

            Assert.IsNotNull(cca.AppConfig.LoggingCallback);
        }

        [TestMethod]
        public void TestConstructor_WithRedirectUri()
        {
            const string RedirectUri = "http://some_redirect_uri/";
            var cca = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                      .WithClientSecret("cats")
                      .WithRedirectUri(RedirectUri).Build();

            Assert.AreEqual(RedirectUri, cca.AppConfig.RedirectUri);
        }

        [TestMethod]
        public void TestConstructor_WithNullRedirectUri()
        {
            var cca = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                      .WithClientSecret("cats")
                      .WithRedirectUri(null).Build();

            Assert.AreEqual(Constants.DefaultConfidentialClientRedirectUri, cca.AppConfig.RedirectUri);
        }

        [TestMethod]
        public void TestConstructor_WithEmptyRedirectUri()
        {
            var cca = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                      .WithClientSecret("cats")
                      .WithRedirectUri(string.Empty).Build();

            Assert.AreEqual(Constants.DefaultConfidentialClientRedirectUri, cca.AppConfig.RedirectUri);
        }

        [TestMethod]
        public void TestConstructor_WithWhitespaceRedirectUri()
        {
            var cca = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                      .WithClientSecret("cats")
                      .WithRedirectUri("      ")
                      .Build();

            Assert.AreEqual(Constants.DefaultConfidentialClientRedirectUri, cca.AppConfig.RedirectUri);
        }

        [TestMethod]
        public void TestConstructor_WithInvalidRedirectUri()
        {
            Assert.ThrowsException<InvalidOperationException>(() =>
                ConfidentialClientApplicationBuilder.Create(TestConstants.ClientId)
                                                    .WithClientSecret("cats")
                                                    .WithRedirectUri("this is not a valid uri")
                                                    .Build());
        }

        [TestMethod]
        public void TestConstructor_WithTenantId()
        {
            const string TenantId = "a_tenant id";
            var cca = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                      .WithClientSecret("cats")
                      .WithTenantId(TenantId)
                      .Build();

            Assert.AreEqual(TenantId, cca.AppConfig.TenantId);
        }

        [TestMethod]
        public void TestConstructor_WithClientSecret()
        {
            const string ClientSecret = "secret value here";
            var cca = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                      .WithClientSecret(ClientSecret)
                      .Build();

            Assert.IsNotNull(cca.AppConfig.ClientSecret);
            Assert.AreEqual(ClientSecret, cca.AppConfig.ClientSecret);
        }

        [TestMethod]
        [DeploymentItem(@"Resources\testCert.crtfile")]
        public void TestConstructor_WithCertificate_X509Certificate2()
        {
            var cert = new X509Certificate2(
                ResourceHelper.GetTestResourceRelativePath("testCert.crtfile"), "passw0rd!");

            var cca = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                      .WithCertificate(cert).Build();

            Assert.IsNotNull(cca.AppConfig.ClientCredentialCertificate);
        }

        [TestMethod]
        [DeploymentItem(@"Resources\valid_cert.cer")]
        public void TestConstructor_WithCertificate_WithoutPrivateKey()
        {
            var cert = new X509Certificate2(
                ResourceHelper.GetTestResourceRelativePath("valid_cert.cer"));

            try
            {
                ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId).WithCertificate(cert).Build();

                Assert.Fail();
            } 
            catch (MsalClientException e)
            {
                Assert.IsNotNull(e);
                Assert.AreEqual(MsalError.CertWithoutPrivateKey, e.ErrorCode);
            }
        }

        [TestMethod]
        [DeploymentItem(@"Resources\testCert.crtfile")]
        public void TestConstructor_WithCertificate_SendX5C()
        {
            var cert = new X509Certificate2(
                ResourceHelper.GetTestResourceRelativePath("testCert.crtfile"), "passw0rd!");

            var app = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                      .WithCertificate(cert)
                      .Build();

            Assert.IsFalse((app.AppConfig as ApplicationConfiguration).SendX5C);

            app = ConfidentialClientApplicationBuilder
                  .Create(TestConstants.ClientId)
                  .WithCertificate(cert, true)
                  .Build();

            Assert.IsTrue((app.AppConfig as ApplicationConfiguration).SendX5C);
        }

        [TestMethod]
        [DeploymentItem(@"Resources\testCert.crtfile")]
        public void TestConstructor_WithClientClaims()
        {
            var cert = new X509Certificate2(
                ResourceHelper.GetTestResourceRelativePath("testCert.crtfile"), "passw0rd!");

            var app = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                      .WithClientClaims(cert, TestConstants.s_clientAssertionClaims)
                      .Build();

            AssertValues(true, false);

            app = ConfidentialClientApplicationBuilder
                .Create(TestConstants.ClientId)
                .WithClientClaims(cert, TestConstants.s_clientAssertionClaims, false)
                .Build();

            AssertValues(false, false);

            app = ConfidentialClientApplicationBuilder
                .Create(TestConstants.ClientId)
                .WithClientClaims(cert, TestConstants.s_clientAssertionClaims, false, true)
                .Build();

            AssertValues(false, true);

            app = ConfidentialClientApplicationBuilder
                .Create(TestConstants.ClientId)
                .WithClientClaims(cert, TestConstants.s_clientAssertionClaims, sendX5C: true)
                .Build();

            AssertValues(true, true);

            void AssertValues(bool expectedMergeWithDefaultClaims, bool expectedSendX5C)
            {
                Assert.AreEqual(expectedSendX5C, (app.AppConfig as ApplicationConfiguration).SendX5C);
                Assert.AreEqual(expectedMergeWithDefaultClaims, (app.AppConfig as ApplicationConfiguration).MergeWithDefaultClaims);
                Assert.IsNotNull((app.AppConfig as ApplicationConfiguration).ClientCredentialCertificate);
                CoreAssert.AssertDictionariesAreEqual(TestConstants.s_clientAssertionClaims, (app.AppConfig as ApplicationConfiguration).ClaimsToSign, StringComparer.Ordinal);
            }
        }

        [TestMethod]
        [DeploymentItem(@"Resources\CustomInstanceMetadata.json")]
        public void TestConstructor_WithValidInstanceDicoveryMetadata()
        {
            string instanceMetadataJson = File.ReadAllText(ResourceHelper.GetTestResourceRelativePath("CustomInstanceMetadata.json"));
            var cca = ConfidentialClientApplicationBuilder.Create(TestConstants.ClientId)
                                                   .WithClientSecret("cats")
                                                   .WithInstanceDiscoveryMetadata(instanceMetadataJson)
                                                   .Build();

            var instanceDiscoveryMetadata = (cca.AppConfig as ApplicationConfiguration).CustomInstanceDiscoveryMetadata;
            Assert.AreEqual(2, instanceDiscoveryMetadata.Metadata.Length);
        }

        [TestMethod]
        [DeploymentItem(@"Resources\CustomInstanceMetadata.json")]
        public void TestConstructor_InstanceMetadata_ValidateAuthority_MutuallyExclusive()
        {
            string instanceMetadataJson = File.ReadAllText(ResourceHelper.GetTestResourceRelativePath("CustomInstanceMetadata.json"));
            var ex = AssertException.Throws<MsalClientException>(() => ConfidentialClientApplicationBuilder.Create(TestConstants.ClientId)
                                                  .WithInstanceDiscoveryMetadata(instanceMetadataJson)
                                                  .WithClientSecret("cats")
                                                  .WithAuthority("https://some.authority/bogus/", true)
                                                  .Build());
            Assert.AreEqual(ex.ErrorCode, MsalError.ValidateAuthorityOrCustomMetadata);
        }

        [TestMethod]
        public void TestConstructor_BadInstanceMetadata()
        {
            var ex = AssertException.Throws<MsalClientException>(() => ConfidentialClientApplicationBuilder.Create(TestConstants.ClientId)
                                                  .WithInstanceDiscoveryMetadata("{bad_json_metadata")
                                                  .WithClientSecret("cats")
                                                  .Build());

            Assert.AreEqual(ex.ErrorCode, MsalError.InvalidUserInstanceMetadata);
        }

        [DataTestMethod]
        [DataRow(true)]
        [DataRow(false)]
        [DataRow(null)] // Not specified, default is true
        public void TestConstructor_WithLegacyCacheCompatibility(bool? isLegacyCacheCompatibilityEnabled)
        {
            var builder = ConfidentialClientApplicationBuilder
                      .Create(TestConstants.ClientId)
                      .WithClientSecret(TestConstants.ClientSecret);

            if (isLegacyCacheCompatibilityEnabled.HasValue)
            {
                builder.WithLegacyCacheCompatibility(isLegacyCacheCompatibilityEnabled.Value);
            } else
            {
                isLegacyCacheCompatibilityEnabled = true;
            }
                      
            var cca = builder.Build();

            Assert.AreEqual(isLegacyCacheCompatibilityEnabled, cca.AppConfig.LegacyCacheCompatibilityEnabled);
        }

        [DataTestMethod]
        [DataRow(true)]
        [DataRow(false)]
        [DataRow(null)] // Not specified, default is true
        public void TestConstructor_WithLegacyCacheCompatibility_WithOptions(bool? isLegacyCacheCompatibilityEnabled)
        {
            var options = CreateConfidentialClientApplicationOptions();
            
            if (isLegacyCacheCompatibilityEnabled.HasValue)
            {
                options.LegacyCacheCompatibilityEnabled = isLegacyCacheCompatibilityEnabled.Value;
            }
            else
            {
                isLegacyCacheCompatibilityEnabled = true;
            }

            var cca = ConfidentialClientApplicationBuilder
                      .CreateWithApplicationOptions(options)
                      .Build();

            Assert.AreEqual(isLegacyCacheCompatibilityEnabled, cca.AppConfig.LegacyCacheCompatibilityEnabled);
        }
    }
}
